@@ -1,16 +1,12 @@
-# 1 "Antialiasing.hs_pre"
-# 1 "<built-in>"
-# 1 "<command-line>"
-# 1 "Antialiasing.hs_pre"
 --------------------------------------------------------------------------------
 -- |
--- Module : Graphics.Rendering.OpenGL.GL.Antialiasing
--- Copyright : (c) Sven Panne 2002-2006
--- License : BSD-style (see the file libraries/OpenGL/LICENSE)
+-- Module      :  Graphics.Rendering.OpenGL.GL.Antialiasing
+-- Copyright   :  (c) Sven Panne 2002-2013
+-- License     :  BSD3
 --
--- Maintainer : sven.panne@aedion.de
--- Stability : stable
--- Portability : portable
+-- Maintainer  :  Sven Panne <svenpanne@gmail.com>
+-- Stability   :  stable
+-- Portability :  portable
 --
 -- This module corresponds to section 3.2 (Antialiasing) of the OpenGL 2.1
 -- specs.
@@ -21,13 +17,10 @@ module Graphics.Rendering.OpenGL.GL.Antialiasing (
    sampleBuffers, samples, multisample, subpixelBits
 ) where
 
-import Graphics.Rendering.OpenGL.GL.Capability (
-   EnableCap(CapMultisample), makeCapability )
-import Graphics.Rendering.OpenGL.GL.BasicTypes ( GLsizei, Capability )
-import Graphics.Rendering.OpenGL.GL.QueryUtils (
-   GetPName(GetSampleBuffers,GetSamples,GetSubpixelBits), getSizei1 )
-import Graphics.Rendering.OpenGL.GL.StateVar (
-   GettableStateVar, makeGettableStateVar, StateVar )
+import Graphics.Rendering.OpenGL.GL.StateVar
+import Graphics.Rendering.OpenGL.GL.Capability
+import Graphics.Rendering.OpenGL.GL.QueryUtils
+import Graphics.Rendering.OpenGL.Raw
 
 --------------------------------------------------------------------------------
 
@@ -43,5 +36,5 @@ multisample = makeCapability CapMultisample
 subpixelBits :: GettableStateVar GLsizei
 subpixelBits = antialiasingInfo GetSubpixelBits
 
-antialiasingInfo :: GetPName -> GettableStateVar GLsizei
+antialiasingInfo :: GetPName1I p => p -> GettableStateVar GLsizei
 antialiasingInfo = makeGettableStateVar . getSizei1 id
