@@ -1,22 +1,20 @@
-# 1 "CoordTrans.hs_pre"
-# 1 "<built-in>"
-# 1 "<command-line>"
-# 1 "CoordTrans.hs_pre"
 --------------------------------------------------------------------------------
 -- |
--- Module : Graphics.Rendering.OpenGL.GL.CoordTrans
--- Copyright : (c) Sven Panne 2002-2006
--- License : BSD-style (see the file libraries/OpenGL/LICENSE)
+-- Module      :  Graphics.Rendering.OpenGL.GL.CoordTrans
+-- Copyright   :  (c) Sven Panne 2002-2013
+-- License     :  BSD3
 --
--- Maintainer : sven.panne@aedion.de
--- Stability : stable
--- Portability : portable
+-- Maintainer  :  Sven Panne <svenpanne@gmail.com>
+-- Stability   :  stable
+-- Portability :  portable
 --
 -- This module corresponds to section 2.11 (Coordinate Transformations) of the
 -- OpenGL 2.1 specs.
 --
 --------------------------------------------------------------------------------
 
+{-# LANGUAGE TypeSynonymInstances #-}
+
 module Graphics.Rendering.OpenGL.GL.CoordTrans (
    -- * Controlling the Viewport
    depthRange,
@@ -24,9 +22,8 @@ module Graphics.Rendering.OpenGL.GL.CoordTrans (
 
    -- * Matrices
    MatrixMode(..), matrixMode,
-   Vector2(..), Vector3(..),
    MatrixOrder(..), MatrixComponent(rotate,translate,scale), Matrix(..),
-   currentMatrix, matrix, multMatrix, GLmatrix, loadIdentity,
+   matrix, multMatrix, GLmatrix, loadIdentity,
    ortho, frustum, depthClamp,
    activeTexture,
    preservingMatrix, unsafePreservingMatrix,
@@ -39,53 +36,21 @@ module Graphics.Rendering.OpenGL.GL.CoordTrans (
    Plane(..), TextureCoordName(..), TextureGenMode(..), textureGenMode
 ) where
 
-import Foreign.ForeignPtr ( ForeignPtr, mallocForeignPtrArray, withForeignPtr )
-import Foreign.Marshal.Alloc ( alloca )
-import Foreign.Marshal.Array ( allocaArray, peekArray, pokeArray, withArray )
-import Foreign.Marshal.Utils ( with )
-import Foreign.Ptr ( Ptr, castPtr )
-import Foreign.Storable ( Storable(..) )
-import Graphics.Rendering.OpenGL.GL.Capability (
-   EnableCap(CapRescaleNormal, CapNormalize,CapDepthClamp,
-             CapTextureGenS, CapTextureGenT,
-             CapTextureGenR, CapTextureGenQ),
-   makeCapability, makeStateVarMaybe )
-import Graphics.Rendering.OpenGL.GL.BasicTypes (
-   GLenum, GLint, GLsizei, GLfloat, GLdouble, GLclampd, Capability(..) )
-import Graphics.Rendering.OpenGL.GL.Exception ( bracket, unsafeBracket_ )
-import Graphics.Rendering.OpenGL.GL.Extensions (
-   FunPtr, unsafePerformIO, Invoker, getProcAddress )
-import Graphics.Rendering.OpenGL.GL.PeekPoke (
-   peek1, peek2, peek3, peek4, poke2, poke3, poke4 )
-import Graphics.Rendering.OpenGL.GL.QueryUtils (
-   GetPName(GetDepthRange,GetViewport,GetMaxViewportDims,GetMatrixMode,
-            GetModelviewMatrix,GetProjectionMatrix,
-            GetTextureMatrix, GetColorMatrix,GetMatrixPalette,GetActiveTexture,
-            GetCurrentMatrixStackDepth,GetModelviewStackDepth,
-            GetMaxModelviewStackDepth,GetProjectionStackDepth,
-            GetMaxProjectionStackDepth,GetTextureStackDepth,
-            GetMaxTextureStackDepth,GetColorMatrixStackDepth,
-            GetMaxColorMatrixStackDepth,GetMaxMatrixPaletteStackDepth),
-   getInteger2, getInteger4, getEnum1, getSizei1, getFloatv,
-   getDouble2, getDoublev, modelviewIndexToEnum, modelviewEnumToIndex )
-import Graphics.Rendering.OpenGL.GL.StateVar (
-   GettableStateVar, makeGettableStateVar, HasGetter(get), HasSetter(($=)),
-   StateVar, makeStateVar )
-import Graphics.Rendering.OpenGL.GL.Texturing.TextureUnit (
-   marshalTextureUnit, unmarshalTextureUnit )
-import Graphics.Rendering.OpenGL.GL.VertexSpec ( TextureUnit )
-import Graphics.Rendering.OpenGL.GLU.ErrorsInternal (
-   recordInvalidEnum, recordInvalidValue )
-
---------------------------------------------------------------------------------
-
-# 1 "../../../../include/HsOpenGLExt.h" 1
-# 79 "CoordTrans.hs_pre" 2
-# 1 "../../../../include/HsOpenGLTypes.h" 1
-# 18 "../../../../include/HsOpenGLTypes.h"
-# 1 "../../../../include/HsOpenGLConfig.h" 1
-# 19 "../../../../include/HsOpenGLTypes.h" 2
-# 80 "CoordTrans.hs_pre" 2
+import Foreign.ForeignPtr
+import Foreign.Marshal.Alloc
+import Foreign.Marshal.Array
+import Foreign.Marshal.Utils
+import Foreign.Ptr
+import Foreign.Storable
+import Graphics.Rendering.OpenGL.GL.StateVar
+import Graphics.Rendering.OpenGL.GL.Tensor
+import Graphics.Rendering.OpenGL.GL.Capability
+import Graphics.Rendering.OpenGL.GL.Exception
+import Graphics.Rendering.OpenGL.GL.PeekPoke
+import Graphics.Rendering.OpenGL.GL.QueryUtils
+import Graphics.Rendering.OpenGL.GL.Texturing.TextureUnit
+import Graphics.Rendering.OpenGL.GLU.ErrorsInternal
+import Graphics.Rendering.OpenGL.Raw
 
 --------------------------------------------------------------------------------
 
@@ -104,10 +69,7 @@ import Graphics.Rendering.OpenGL.GLU.ErrorsInternal (
 -- mappings such as (1, 0) are acceptable.
 
 depthRange :: StateVar (GLclampd, GLclampd)
-depthRange = makeStateVar (getDouble2 (,) GetDepthRange) (uncurry glDepthRange)
-
-foreign import unsafe "glDepthRange" glDepthRange ::
-   GLclampd -> GLclampd -> IO ()
+depthRange = makeStateVar (getClampd2 (,) GetDepthRange) (uncurry glDepthRange)
 
 --------------------------------------------------------------------------------
 
@@ -129,7 +91,7 @@ data Size = Size !GLsizei !GLsizei
 -- Let (/xnd/, /ynd/) be normalized device coordinates. Then the window
 -- coordinates (/xw/, /yw/) are computed as follows:
 --
--- /xw/ = (/xnd/ + 1) (/width/ \/ 2) + /x/
+-- /xw/ = (/xnd/ + 1) (/width/  \/ 2) + /x/
 --
 -- /yw/ = (/ynd/ + 1) (/heigth/ \/ 2) + /y/
 --
@@ -141,68 +103,65 @@ viewport = makeStateVar (getInteger4 makeVp GetViewport)
                         (\(Position x y, Size w h) -> glViewport x y w h)
    where makeVp x y w h = (Position x y, Size (fromIntegral w) (fromIntegral h))
 
-foreign import unsafe "glViewport" glViewport ::
-   GLint -> GLint -> GLsizei -> GLsizei -> IO ()
-
 -- | The implementation-dependent maximum viewport width and height.
 
 maxViewportDims :: GettableStateVar Size
-maxViewportDims = makeGettableStateVar (getInteger2 Size GetMaxViewportDims)
+maxViewportDims = makeGettableStateVar (getSizei2 Size GetMaxViewportDims)
 
 --------------------------------------------------------------------------------
 
 -- | A matrix stack.
 
 data MatrixMode =
-     Modelview GLsizei -- ^ The modelview matrix stack of the specified vertex unit.
-   | Projection -- ^ The projection matrix stack.
-   | Texture -- ^ The texture matrix stack.
-   | Color -- ^ The color matrix stack.
-   | MatrixPalette -- ^ The matrix palette stack.
+     Modelview GLsizei  -- ^ The modelview matrix stack of the specified vertex unit.
+   | Projection         -- ^ The projection matrix stack.
+   | Texture            -- ^ The texture matrix stack.
+   | Color              -- ^ The color matrix stack.
+   | MatrixPalette      -- ^ The matrix palette stack.
    deriving ( Eq, Ord, Show )
 
 marshalMatrixMode :: MatrixMode -> Maybe GLenum
 marshalMatrixMode x = case x of
    Modelview i -> modelviewIndexToEnum i
-   Projection -> Just 0x1701
-   Texture -> Just 0x1702
-   Color -> Just 0x1800
-   MatrixPalette -> Just 0x8840
+   Projection -> Just gl_PROJECTION
+   Texture -> Just gl_TEXTURE
+   Color -> Just gl_COLOR
+   MatrixPalette -> Just gl_MATRIX_PALETTE
 
 unmarshalMatrixMode :: GLenum -> MatrixMode
 unmarshalMatrixMode x
-   | x == 0x1701 = Projection
-   | x == 0x1702 = Texture
-   | x == 0x1800 = Color
-   | x == 0x8840 = MatrixPalette
+   | x == gl_PROJECTION = Projection
+   | x == gl_TEXTURE = Texture
+   | x == gl_COLOR = Color
+   | x == gl_MATRIX_PALETTE = MatrixPalette
    | otherwise =
         case modelviewEnumToIndex x of
            Just i -> Modelview i
            Nothing -> error ("unmarshalMatrixMode: illegal value " ++ show x)
 
-matrixModeToGetMatrix :: MatrixMode -> GetPName
+matrixModeToGetMatrix :: MatrixMode -> PNameMatrix
 matrixModeToGetMatrix x = case x of
-   Modelview _ -> GetModelviewMatrix -- ???
-   Projection -> GetProjectionMatrix
-   Texture -> GetTextureMatrix
-   Color -> GetColorMatrix
+   Modelview _   -> GetModelviewMatrix -- ???
+   Projection    -> GetProjectionMatrix
+   Texture       -> GetTextureMatrix
+   Color         -> GetColorMatrix
    MatrixPalette -> GetMatrixPalette
 
-matrixModeToGetStackDepth :: MatrixMode -> GetPName
-matrixModeToGetStackDepth x = case x of
-   Modelview _ -> GetModelviewStackDepth
-   Projection -> GetProjectionStackDepth
-   Texture -> GetTextureStackDepth
-   Color -> GetColorMatrixStackDepth
+matrixModeToGetStackDepth :: MatrixMode -> PName1I
+matrixModeToGetStackDepth x =  case x of
+   Modelview _   -> GetModelviewStackDepth
+   Projection    -> GetProjectionStackDepth
+   Texture       -> GetTextureStackDepth
+   Color         -> GetColorMatrixStackDepth
    MatrixPalette -> error "matrixModeToGetStackDepth: impossible"
 
-matrixModeToGetMaxStackDepth :: MatrixMode -> GetPName
+matrixModeToGetMaxStackDepth :: MatrixMode -> PName1I
 matrixModeToGetMaxStackDepth x = case x of
-   Modelview _ -> GetMaxModelviewStackDepth
-   Projection -> GetMaxProjectionStackDepth
-   Texture -> GetMaxTextureStackDepth
-   Color -> GetMaxColorMatrixStackDepth
-   MatrixPalette -> GetMaxMatrixPaletteStackDepth
+   Modelview _    -> GetMaxModelviewStackDepth
+   Projection     -> GetMaxProjectionStackDepth
+   Texture        -> GetMaxTextureStackDepth
+   Color          -> GetMaxColorMatrixStackDepth
+   MatrixPalette  -> GetMaxMatrixPaletteStackDepth
 
 --------------------------------------------------------------------------------
 
@@ -214,30 +173,6 @@ matrixMode =
    makeStateVar (getEnum1 unmarshalMatrixMode GetMatrixMode)
                 (maybe recordInvalidValue glMatrixMode . marshalMatrixMode)
 
-foreign import unsafe "glMatrixMode" glMatrixMode :: GLenum -> IO ()
-
---------------------------------------------------------------------------------
-
-data Vector2 a = Vector2 !a !a
-   deriving ( Eq, Ord, Show )
-
-instance Storable a => Storable (Vector2 a) where
-   sizeOf ~(Vector2 x _) = 2 * sizeOf x
-   alignment ~(Vector2 x _) = alignment x
-   peek = peek2 Vector2 . castPtr
-   poke ptr (Vector2 x y) = poke2 (castPtr ptr) x y
-
---------------------------------------------------------------------------------
-
-data Vector3 a = Vector3 !a !a !a
-   deriving ( Eq, Ord, Show )
-
-instance Storable a => Storable (Vector3 a) where
-   sizeOf ~(Vector3 x _ _) = 3 * sizeOf x
-   alignment ~(Vector3 x _ _) = alignment x
-   peek = peek3 Vector3 . castPtr
-   poke ptr (Vector3 x y z) = poke3 (castPtr ptr) x y z
-
 --------------------------------------------------------------------------------
 
 data MatrixOrder = ColumnMajor | RowMajor
@@ -246,7 +181,7 @@ data MatrixOrder = ColumnMajor | RowMajor
 --------------------------------------------------------------------------------
 
 class Storable c => MatrixComponent c where
-   getMatrix :: GetPName -> Ptr c -> IO ()
+   getMatrix :: GetPNameMatrix p => p -> Ptr c -> IO ()
    loadMatrix :: Ptr c -> IO ()
    loadTransposeMatrix :: Ptr c -> IO ()
    multMatrix_ :: Ptr c -> IO ()
@@ -255,51 +190,28 @@ class Storable c => MatrixComponent c where
    translate :: Vector3 c -> IO ()
    scale :: c -> c -> c -> IO ()
 
-instance MatrixComponent Float where
-   getMatrix = getFloatv
+instance MatrixComponent GLfloat where
+   getMatrix = getMatrixf
    loadMatrix = glLoadMatrixf
-   loadTransposeMatrix = glLoadTransposeMatrixfARB
+   loadTransposeMatrix = glLoadTransposeMatrixf
    multMatrix_ = glMultMatrixf
-   multTransposeMatrix = glMultTransposeMatrixfARB
+   multTransposeMatrix = glMultTransposeMatrixf
    rotate a (Vector3 x y z) = glRotatef a x y z
    translate (Vector3 x y z) = glTranslatef x y z
    scale = glScalef
 
-instance MatrixComponent Double where
-   getMatrix = getDoublev
+instance MatrixComponent GLdouble where
+   getMatrix = getMatrixd
    loadMatrix = glLoadMatrixd
-   loadTransposeMatrix = glLoadTransposeMatrixdARB
+   loadTransposeMatrix = glLoadTransposeMatrixd
    multMatrix_ = glMultMatrixd
-   multTransposeMatrix = glMultTransposeMatrixdARB
+   multTransposeMatrix = glMultTransposeMatrixd
    rotate a (Vector3 x y z) = glRotated a x y z
    translate (Vector3 x y z) = glTranslated x y z
    scale = glScaled
 
 --------------------------------------------------------------------------------
 
-foreign import unsafe "glLoadMatrixf" glLoadMatrixf :: Ptr GLfloat -> IO ()
-foreign import unsafe "glLoadMatrixd" glLoadMatrixd :: Ptr GLdouble -> IO ()
-
-foreign import unsafe "dynamic" dyn_glLoadTransposeMatrixfARB :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (Ptr GLfloat -> IO ()) ; glLoadTransposeMatrixfARB :: (Ptr GLfloat -> IO ()) ; glLoadTransposeMatrixfARB = dyn_glLoadTransposeMatrixfARB ptr_glLoadTransposeMatrixfARB ; ptr_glLoadTransposeMatrixfARB :: FunPtr a ; ptr_glLoadTransposeMatrixfARB = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_ARB_transpose_matrix or OpenGL 1.3") ("_entry")) ;
-foreign import unsafe "dynamic" dyn_glLoadTransposeMatrixdARB :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (Ptr GLdouble -> IO ()) ; glLoadTransposeMatrixdARB :: (Ptr GLdouble -> IO ()) ; glLoadTransposeMatrixdARB = dyn_glLoadTransposeMatrixdARB ptr_glLoadTransposeMatrixdARB ; ptr_glLoadTransposeMatrixdARB :: FunPtr a ; ptr_glLoadTransposeMatrixdARB = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_ARB_transpose_matrix or OpenGL 1.3") ("_entry")) ;
-
-foreign import unsafe "glMultMatrixf" glMultMatrixf :: Ptr GLfloat -> IO ()
-foreign import unsafe "glMultMatrixd" glMultMatrixd :: Ptr GLdouble -> IO ()
-
-foreign import unsafe "dynamic" dyn_glMultTransposeMatrixfARB :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (Ptr GLfloat -> IO ()) ; glMultTransposeMatrixfARB :: (Ptr GLfloat -> IO ()) ; glMultTransposeMatrixfARB = dyn_glMultTransposeMatrixfARB ptr_glMultTransposeMatrixfARB ; ptr_glMultTransposeMatrixfARB :: FunPtr a ; ptr_glMultTransposeMatrixfARB = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_ARB_transpose_matrix or OpenGL 1.3") ("_entry")) ;
-foreign import unsafe "dynamic" dyn_glMultTransposeMatrixdARB :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (Ptr GLdouble -> IO ()) ; glMultTransposeMatrixdARB :: (Ptr GLdouble -> IO ()) ; glMultTransposeMatrixdARB = dyn_glMultTransposeMatrixdARB ptr_glMultTransposeMatrixdARB ; ptr_glMultTransposeMatrixdARB :: FunPtr a ; ptr_glMultTransposeMatrixdARB = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_ARB_transpose_matrix or OpenGL 1.3") ("_entry")) ;
-
-foreign import unsafe "glRotatef" glRotatef :: GLfloat -> GLfloat -> GLfloat -> GLfloat -> IO ()
-foreign import unsafe "glRotated" glRotated :: GLdouble -> GLdouble -> GLdouble -> GLdouble -> IO ()
-
-foreign import unsafe "glTranslatef" glTranslatef :: GLfloat -> GLfloat -> GLfloat -> IO ()
-foreign import unsafe "glTranslated" glTranslated :: GLdouble -> GLdouble -> GLdouble -> IO ()
-
-foreign import unsafe "glScalef" glScalef :: GLfloat -> GLfloat -> GLfloat -> IO ()
-foreign import unsafe "glScaled" glScaled :: GLdouble -> GLdouble -> GLdouble -> IO ()
-
---------------------------------------------------------------------------------
-
 class Matrix m where
    -- | Create a new matrix of the given order (containing undefined elements)
    -- and call the action to fill it with 4x4 elements.
@@ -331,17 +243,13 @@ class Matrix m where
       withMatrix mat $ \order p ->
         if desiredOrder == order
            then peekArray 16 p
-           else mapM (peekElemOff p) [ 0, 4, 8, 12,
-                                       1, 5, 9, 13,
+           else mapM (peekElemOff p) [ 0, 4,  8, 12,
+                                       1, 5,  9, 13,
                                        2, 6, 10, 14,
                                        3, 7, 11, 15 ]
 
 --------------------------------------------------------------------------------
 
-{-# DEPRECATED currentMatrix "use `matrix' instead" #-}
-currentMatrix :: (Matrix m, MatrixComponent c) => StateVar (m c)
-currentMatrix = matrix Nothing
-
 matrix :: (Matrix m, MatrixComponent c) => Maybe MatrixMode -> StateVar (m c)
 matrix maybeMode =
    makeStateVar
@@ -354,7 +262,7 @@ withMatrixMode mode act =
       matrixMode $= mode
       act
 
-getMatrix' :: (Matrix m, MatrixComponent c) => GetPName -> IO (m c)
+getMatrix' :: (Matrix m, MatrixComponent c) => PNameMatrix -> IO (m c)
 getMatrix' = withNewMatrix ColumnMajor . getMatrix
 
 setMatrix :: (Matrix m, MatrixComponent c) => m c -> IO ()
@@ -362,27 +270,20 @@ setMatrix mat =
    withMatrix mat $ \order ->
       case order of
          ColumnMajor -> loadMatrix
-         RowMajor -> loadTransposeMatrix
+         RowMajor    -> loadTransposeMatrix
 
 multMatrix :: (Matrix m, MatrixComponent c) => m c -> IO ()
 multMatrix mat =
    withMatrix mat $ \order ->
       case order of
          ColumnMajor -> multMatrix_
-         RowMajor -> multTransposeMatrix
+         RowMajor    -> multTransposeMatrix
 
 --------------------------------------------------------------------------------
 
 data GLmatrix a = GLmatrix MatrixOrder (ForeignPtr a)
-
-
-
-
-
-
    deriving ( Eq, Ord, Show )
 
-
 instance Matrix GLmatrix where
    withNewMatrix order f = do
       fp <- mallocForeignPtrArray 16
@@ -393,15 +294,16 @@ instance Matrix GLmatrix where
 
 --------------------------------------------------------------------------------
 
-foreign import unsafe "glLoadIdentity" loadIdentity :: IO ()
+loadIdentity :: IO ()
+loadIdentity = glLoadIdentity
 
 --------------------------------------------------------------------------------
 
-foreign import unsafe "glOrtho" ortho ::
-   GLdouble -> GLdouble -> GLdouble -> GLdouble -> GLdouble -> GLdouble -> IO ()
+ortho :: GLdouble -> GLdouble -> GLdouble -> GLdouble -> GLdouble -> GLdouble -> IO ()
+ortho = glOrtho
 
-foreign import unsafe "glFrustum" frustum ::
-   GLdouble -> GLdouble -> GLdouble -> GLdouble -> GLdouble -> GLdouble -> IO ()
+frustum :: GLdouble -> GLdouble -> GLdouble -> GLdouble -> GLdouble -> GLdouble -> IO ()
+frustum = glFrustum
 
 --------------------------------------------------------------------------------
 
@@ -412,9 +314,7 @@ depthClamp = makeCapability CapDepthClamp
 
 activeTexture :: StateVar TextureUnit
 activeTexture = makeStateVar (getEnum1 unmarshalTextureUnit GetActiveTexture)
-                             (glActiveTextureARB . marshalTextureUnit)
-
-foreign import unsafe "dynamic" dyn_glActiveTextureARB :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLenum -> IO ()) ; glActiveTextureARB :: (GLenum -> IO ()) ; glActiveTextureARB = dyn_glActiveTextureARB ptr_glActiveTextureARB ; ptr_glActiveTextureARB :: FunPtr a ; ptr_glActiveTextureARB = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_ARB_multitexture or OpenGL 1.3") ("_entry")) ;
+                             (glActiveTexture . marshalTextureUnit)
 
 --------------------------------------------------------------------------------
 
@@ -439,10 +339,6 @@ preservingMatrixMode = bracket (getEnum1 id GetMatrixMode) glMatrixMode . const
 unsafePreservingMatrix :: IO a -> IO a
 unsafePreservingMatrix = unsafeBracket_ glPushMatrix glPopMatrix
 
-foreign import unsafe "glPushMatrix" glPushMatrix :: IO ()
-
-foreign import unsafe "glPopMatrix" glPopMatrix :: IO ()
-
 --------------------------------------------------------------------------------
 
 stackDepth :: Maybe MatrixMode -> GettableStateVar GLsizei
@@ -482,10 +378,10 @@ data Plane a = Plane !a !a !a !a
    deriving ( Eq, Ord, Show )
 
 instance Storable a => Storable (Plane a) where
-   sizeOf ~(Plane a _ _ _) = 4 * sizeOf a
+   sizeOf    ~(Plane a _ _ _) = 4 * sizeOf a
    alignment ~(Plane a _ _ _) = alignment a
-   peek = peek4 Plane . castPtr
-   poke ptr (Plane a b c d) = poke4 (castPtr ptr) a b c d
+   peek                       = peek4 Plane . castPtr
+   poke ptr   (Plane a b c d) = poke4 (castPtr ptr) a b c d
 
 --------------------------------------------------------------------------------
 
@@ -498,10 +394,10 @@ data TextureCoordName =
 
 marshalTextureCoordName :: TextureCoordName -> GLenum
 marshalTextureCoordName x = case x of
-   S -> 0x2000
-   T -> 0x2001
-   R -> 0x2002
-   Q -> 0x2003
+   S -> gl_S
+   T -> gl_T
+   R -> gl_R
+   Q -> gl_Q
 
 --------------------------------------------------------------------------------
 
@@ -512,9 +408,9 @@ data TextureGenParameter =
 
 marshalTextureGenParameter :: TextureGenParameter -> GLenum
 marshalTextureGenParameter x = case x of
-   TextureGenMode -> 0x2500
-   ObjectPlane -> 0x2501
-   EyePlane -> 0x2502
+   TextureGenMode -> gl_TEXTURE_GEN_MODE
+   ObjectPlane -> gl_OBJECT_PLANE
+   EyePlane -> gl_EYE_PLANE
 
 --------------------------------------------------------------------------------
 
@@ -526,26 +422,27 @@ data TextureGenMode' =
    | ReflectionMap'
 
 marshalTextureGenMode' :: TextureGenMode' -> GLint
-marshalTextureGenMode' x = case x of
-   EyeLinear' -> 0x2400
-   ObjectLinear' -> 0x2401
-   SphereMap' -> 0x2402
-   NormalMap' -> 0x8511
-   ReflectionMap' -> 0x8512
+marshalTextureGenMode' x = fromIntegral $ case x of
+   EyeLinear' -> gl_EYE_LINEAR
+   ObjectLinear' -> gl_OBJECT_LINEAR
+   SphereMap' -> gl_SPHERE_MAP
+   NormalMap' -> gl_NORMAL_MAP
+   ReflectionMap' -> gl_REFLECTION_MAP
 
 unmarshalTextureGenMode' :: GLint -> TextureGenMode'
 unmarshalTextureGenMode' x
-   | x == 0x2400 = EyeLinear'
-   | x == 0x2401 = ObjectLinear'
-   | x == 0x2402 = SphereMap'
-   | x == 0x8511 = NormalMap'
-   | x == 0x8512 = ReflectionMap'
+   | y == gl_EYE_LINEAR = EyeLinear'
+   | y == gl_OBJECT_LINEAR = ObjectLinear'
+   | y == gl_SPHERE_MAP = SphereMap'
+   | y == gl_NORMAL_MAP = NormalMap'
+   | y == gl_REFLECTION_MAP = ReflectionMap'
    | otherwise = error ("unmarshalTextureGenMode': illegal value " ++ show x)
+   where y = fromIntegral x
 
 --------------------------------------------------------------------------------
 
 data TextureGenMode =
-     EyeLinear (Plane GLdouble)
+     EyeLinear    (Plane GLdouble)
    | ObjectLinear (Plane GLdouble)
    | SphereMap
    | NormalMap
@@ -554,11 +451,11 @@ data TextureGenMode =
 
 marshalTextureGenMode :: TextureGenMode -> GLint
 marshalTextureGenMode = marshalTextureGenMode' . convertMode
-   where convertMode (EyeLinear _) = EyeLinear'
+   where convertMode (EyeLinear    _) = EyeLinear'
          convertMode (ObjectLinear _) = ObjectLinear'
-         convertMode SphereMap = SphereMap'
-         convertMode NormalMap = NormalMap'
-         convertMode ReflectionMap = ReflectionMap'
+         convertMode SphereMap        = SphereMap'
+         convertMode NormalMap        = NormalMap'
+         convertMode ReflectionMap    = ReflectionMap'
 
 --------------------------------------------------------------------------------
 
@@ -576,7 +473,7 @@ textureGenMode coord =
       (\mode -> do
          setMode coord mode
          case mode of
-            EyeLinear plane -> setPlane coord EyePlane plane
+            EyeLinear    plane -> setPlane coord EyePlane    plane
             ObjectLinear plane -> setPlane coord ObjectPlane plane
             _ -> return ())
 
@@ -598,36 +495,24 @@ getMode coord = alloca $ \buf -> do
                  buf
    peek1 unmarshalTextureGenMode' buf
 
-foreign import unsafe "glGetTexGeniv" glGetTexGeniv ::
-   GLenum -> GLenum -> Ptr GLint -> IO ()
-
 setMode :: TextureCoordName -> TextureGenMode -> IO ()
 setMode coord mode =
    glTexGeni (marshalTextureCoordName coord)
              (marshalTextureGenParameter TextureGenMode)
              (marshalTextureGenMode mode)
 
-foreign import unsafe "glTexGeni" glTexGeni ::
-    GLenum -> GLenum -> GLint -> IO ()
-
 --------------------------------------------------------------------------------
 
 getPlane :: TextureCoordName -> TextureGenParameter -> IO (Plane GLdouble)
 getPlane coord param = alloca $ \planeBuffer -> do
    glGetTexGendv (marshalTextureCoordName coord)
                  (marshalTextureGenParameter param)
-                 planeBuffer
+                 (castPtr planeBuffer)
    peek planeBuffer
 
-foreign import unsafe "glGetTexGendv" glGetTexGendv ::
-   GLenum -> GLenum -> Ptr (Plane GLdouble) -> IO ()
-
 setPlane :: TextureCoordName -> TextureGenParameter -> Plane GLdouble -> IO ()
 setPlane coord param plane =
    with plane $ \planeBuffer ->
       glTexGendv (marshalTextureCoordName coord)
                  (marshalTextureGenParameter param)
-                 planeBuffer
-
-foreign import unsafe "glTexGendv" glTexGendv ::
-   GLenum -> GLenum -> Ptr (Plane GLdouble) -> IO ()
+                 (castPtr planeBuffer)
