@@ -1,16 +1,12 @@
-# 1 "Framebuffer.hs_pre"
-# 1 "<built-in>"
-# 1 "<command-line>"
-# 1 "Framebuffer.hs_pre"
 --------------------------------------------------------------------------------
 -- |
--- Module : Graphics.Rendering.OpenGL.GL.Framebuffer
--- Copyright : (c) Sven Panne 2002-2006
--- License : BSD-style (see the file libraries/OpenGL/LICENSE)
+-- Module      :  Graphics.Rendering.OpenGL.GL.Framebuffer
+-- Copyright   :  (c) Sven Panne 2002-2013
+-- License     :  BSD3
 --
--- Maintainer : sven.panne@aedion.de
--- Stability : stable
--- Portability : portable
+-- Maintainer  :  Sven Panne <svenpanne@gmail.com>
+-- Stability   :  stable
+-- Portability :  portable
 --
 -- This module corresponds to section 4.2 (Whole Framebuffer Operations) of the
 -- OpenGL 2.1 specs.
@@ -20,13 +16,13 @@
 module Graphics.Rendering.OpenGL.GL.Framebuffer (
    -- * Querying the Buffer Configuration
    auxBuffers, doubleBuffer, stereoBuffer,
-   rgbaBits, stencilBits, depthBits, accumBits,
+   rgbaBits, stencilBits, depthBits, accumBits, rgbaSignedComponents,
 
    -- * Selecting a Buffer for Writing
-   BufferMode(..), drawBuffer, drawBuffers, maxDrawBuffers,
+   DrawBufferIndex, BufferMode(..), drawBuffer, drawBuffers, drawBufferi, maxDrawBuffers,
 
    -- * Fine Control of Buffer Updates
-   indexMask, colorMask, stencilMask, stencilMaskSeparate, depthMask,
+   indexMask, colorMask, colorMaski, stencilMask, stencilMaskSeparate, depthMask,
 
    -- * Clearing the Buffers
    ClearBuffer(..), clear,
@@ -36,50 +32,19 @@ module Graphics.Rendering.OpenGL.GL.Framebuffer (
    AccumOp(..), accum
 ) where
 
-import Control.Monad ( liftM4 )
-import Data.List ( genericLength )
-import Data.Maybe ( isJust, fromJust )
-import Foreign.Marshal.Array ( withArray )
-import Foreign.Ptr ( Ptr )
-import Graphics.Rendering.OpenGL.GL.BufferMode (
-   BufferMode(..), marshalBufferMode, unmarshalBufferMode )
-import Graphics.Rendering.OpenGL.GL.Capability (
-   marshalCapability, unmarshalCapability )
-import Graphics.Rendering.OpenGL.GL.BasicTypes (
-   GLboolean, GLenum, GLsizei, GLint, GLuint, GLbitfield, GLfloat, GLclampf,
-   GLclampd,
-
-
-
-   Capability
-
-   )
-import Graphics.Rendering.OpenGL.GL.Extensions (
-   FunPtr, unsafePerformIO, Invoker, getProcAddress )
-import Graphics.Rendering.OpenGL.GL.Face ( Face, marshalFace )
-import Graphics.Rendering.OpenGL.GL.GLboolean ( unmarshalGLboolean )
-import Graphics.Rendering.OpenGL.GL.QueryUtils (
-   GetPName(GetAuxBuffers,GetDoublebuffer,GetStereo,GetRedBits,GetGreenBits,
-            GetBlueBits,GetAlphaBits,GetStencilBits,GetDepthBits,
-            GetAccumRedBits,GetAccumGreenBits,GetAccumBlueBits,
-            GetAccumAlphaBits,GetDrawBuffer,GetIndexWritemask,GetColorWritemask,
-            GetDepthWritemask,GetStencilWritemask,GetColorClearValue,
-            GetIndexClearValue,GetDepthClearValue,GetStencilClearValue,
-            GetAccumClearValue,GetMaxDrawBuffers,GetDrawBufferN),
-   getInteger1, getBoolean1, getBoolean4, getEnum1, getSizei1, getFloat1,
-   getFloat4, getDouble1 )
-import Graphics.Rendering.OpenGL.GL.StateVar (
-   HasGetter(get), GettableStateVar, makeGettableStateVar,
-   SettableStateVar, makeSettableStateVar,
-   StateVar, makeStateVar )
-import Graphics.Rendering.OpenGL.GL.VertexSpec (
-   Color4(Color4), Index1(Index1) )
-import Graphics.Rendering.OpenGL.GLU.ErrorsInternal ( recordInvalidValue )
-
---------------------------------------------------------------------------------
-
-# 1 "../../../../include/HsOpenGLExt.h" 1
-# 78 "Framebuffer.hs_pre" 2
+import Control.Monad
+import Data.List
+import Data.Maybe
+import Foreign.Marshal.Array
+import Graphics.Rendering.OpenGL.GL.BufferMode
+import Graphics.Rendering.OpenGL.GL.Capability
+import Graphics.Rendering.OpenGL.GL.Face
+import Graphics.Rendering.OpenGL.GL.GLboolean
+import Graphics.Rendering.OpenGL.GL.QueryUtils
+import Graphics.Rendering.OpenGL.GL.StateVar
+import Graphics.Rendering.OpenGL.GL.VertexSpec
+import Graphics.Rendering.OpenGL.GLU.ErrorsInternal
+import Graphics.Rendering.OpenGL.Raw
 
 --------------------------------------------------------------------------------
 
@@ -122,6 +87,15 @@ accumBits =
                     (getSizei1 id GetAccumBlueBits)
                     (getSizei1 id GetAccumAlphaBits)
 
+rgbaSignedComponents :: GettableStateVar (Color4 Bool)
+rgbaSignedComponents =
+   makeGettableStateVar $
+      getInteger4 (\r g b a -> Color4 (unmarshalGLboolean r)
+                                      (unmarshalGLboolean g)
+                                      (unmarshalGLboolean b)
+                                      (unmarshalGLboolean a))
+                  GetRGBASignedComponents
+
 --------------------------------------------------------------------------------
 
 -- | When colors are written to the framebuffer, they are written into the color
@@ -145,11 +119,10 @@ drawBuffer =
       (getEnum1 unmarshalBufferMode GetDrawBuffer)
       (maybe recordInvalidValue glDrawBuffer . marshalBufferMode)
 
-foreign import unsafe "glDrawBuffer" glDrawBuffer ::
-   GLenum -> IO ()
-
 --------------------------------------------------------------------------------
 
+type DrawBufferIndex = GLuint
+
 -- | 'drawBuffers' defines the draw buffers to which all fragment colors are
 -- written. The draw buffers being defined correspond in order to the respective
 -- fragment colors. The draw buffer for fragment colors beyond those specified
@@ -187,7 +160,13 @@ setDrawBuffers modes = do
               glDrawBuffers (genericLength ms)
       else recordInvalidValue
 
-foreign import unsafe "dynamic" dyn_glDrawBuffers :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLsizei -> Ptr GLenum -> IO ()) ; glDrawBuffers :: (GLsizei -> Ptr GLenum -> IO ()) ; glDrawBuffers = dyn_glDrawBuffers ptr_glDrawBuffers ; ptr_glDrawBuffers :: FunPtr a ; ptr_glDrawBuffers = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_ARB_draw_buffers or OpenGL 2.0") ("_entry")) ;
+-- | 'drawBufferi' is a fast query function. For indices in the range 0..maxDrawBuffers it's results
+-- are the same as selection the index from the list returned by drawBuffers. Though this function
+-- only uses one gl-function call instead of maxDrawBuffers + 1.
+drawBufferi :: DrawBufferIndex -> GettableStateVar BufferMode
+drawBufferi ind = makeGettableStateVar
+    (getEnum1 unmarshalBufferMode . GetDrawBufferN $ fromIntegral ind)
+
 
 --------------------------------------------------------------------------------
 
@@ -204,7 +183,7 @@ maxDrawBuffers = makeGettableStateVar $ getSizei1 id GetMaxDrawBuffers
 
 -- | Controls the writing of individual bits in the color index buffers. The
 -- least significant /n/ bits of its value, where /n/ is the number of bits in a
--- color index buffer, specify a mask. Where a 1 appears in the mask, it is
+-- color index buffer, specify a mask.  Where a 1 appears in the mask, it is
 -- possible to write to the corresponding bit in the color index buffer (or
 -- buffers). Where a 0 appears, the corresponding bit is write-protected.
 --
@@ -216,8 +195,6 @@ indexMask :: StateVar GLuint
 indexMask =
    makeStateVar (getInteger1 fromIntegral GetIndexWritemask) glIndexMask
 
-foreign import unsafe "glIndexMask" glIndexMask :: GLuint -> IO ()
-
 --------------------------------------------------------------------------------
 
 -- | Controls whether the individual color components in the framebuffer can or
@@ -242,9 +219,18 @@ colorMask =
                                         (marshalCapability g)
                                         (marshalCapability b)
                                         (marshalCapability a))
-
-foreign import unsafe "glColorMask" glColorMask ::
-   GLboolean -> GLboolean -> GLboolean -> GLboolean -> IO ()
+-- | 'colorMaski' is a version of 'colorMask' that only applies to the specified drawbuffer
+colorMaski :: DrawBufferIndex -> StateVar (Color4 Capability)
+colorMaski x = makeStateVar
+      (getBoolean4i (\r g b a -> Color4 (unmarshalCapability r)
+                                       (unmarshalCapability g)
+                                       (unmarshalCapability b)
+                                       (unmarshalCapability a))
+                    GetColorWritemask x)
+      (\(Color4 r g b a) -> glColorMaski (x) (marshalCapability r)
+                                        (marshalCapability g)
+                                        (marshalCapability b)
+                                        (marshalCapability a))
 
 --------------------------------------------------------------------------------
 
@@ -255,8 +241,6 @@ depthMask :: StateVar Capability
 depthMask = makeStateVar (getBoolean1 unmarshalCapability GetDepthWritemask)
                          (glDepthMask . marshalCapability)
 
-foreign import unsafe "glDepthMask" glDepthMask :: GLboolean -> IO ()
-
 --------------------------------------------------------------------------------
 
 -- | Controls the writing of individual bits in the stencil planes. The least
@@ -270,32 +254,28 @@ stencilMask :: StateVar GLuint
 stencilMask =
    makeStateVar (getInteger1 fromIntegral GetStencilWritemask) glStencilMask
 
-foreign import unsafe "glStencilMask" glStencilMask :: GLuint -> IO ()
-
 stencilMaskSeparate :: Face -> SettableStateVar GLuint
 stencilMaskSeparate face =
    makeSettableStateVar $
       glStencilMaskSeparate (marshalFace face)
 
-foreign import unsafe "dynamic" dyn_glStencilMaskSeparate :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLenum -> GLuint -> IO ()) ; glStencilMaskSeparate :: (GLenum -> GLuint -> IO ()) ; glStencilMaskSeparate = dyn_glStencilMaskSeparate ptr_glStencilMaskSeparate ; ptr_glStencilMaskSeparate :: FunPtr a ; ptr_glStencilMaskSeparate = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("OpenGL 2.0") ("_entry")) ;
-
 --------------------------------------------------------------------------------
 
 -- | The buffers which can be cleared with 'clear'.
 
 data ClearBuffer =
-     ColorBuffer -- ^ The buffers currently enabled for color writing.
-   | AccumBuffer -- ^ The accumulation buffer.
+     ColorBuffer   -- ^ The buffers currently enabled for color writing.
+   | AccumBuffer   -- ^ The accumulation buffer.
    | StencilBuffer -- ^ The stencil buffer.
-   | DepthBuffer -- ^ The depth buffer.
+   | DepthBuffer   -- ^ The depth buffer.
    deriving ( Eq, Ord, Show )
 
 marshalClearBuffer :: ClearBuffer -> GLbitfield
 marshalClearBuffer x = case x of
-   ColorBuffer -> 0x4000
-   AccumBuffer -> 0x200
-   StencilBuffer -> 0x400
-   DepthBuffer -> 0x100
+   ColorBuffer -> gl_COLOR_BUFFER_BIT
+   AccumBuffer -> gl_ACCUM_BUFFER_BIT
+   StencilBuffer -> gl_STENCIL_BUFFER_BIT
+   DepthBuffer -> gl_DEPTH_BUFFER_BIT
 
 --------------------------------------------------------------------------------
 
@@ -319,8 +299,6 @@ marshalClearBuffer x = case x of
 clear :: [ClearBuffer] -> IO ()
 clear = glClear . sum . map marshalClearBuffer
 
-foreign import unsafe "glClear" glClear :: GLbitfield -> IO ()
-
 --------------------------------------------------------------------------------
 
 -- | Controls the red, green, blue, and alpha values used by 'clear' to clear
@@ -328,12 +306,9 @@ foreign import unsafe "glClear" glClear :: GLbitfield -> IO ()
 -- [0, 1]. Initially, all values are 0.
 
 clearColor :: StateVar (Color4 GLclampf)
-clearColor = makeStateVar (getFloat4 Color4 GetColorClearValue)
+clearColor = makeStateVar (getClampf4 Color4 GetColorClearValue)
                           (\(Color4 r g b a) -> glClearColor r g b a)
 
-foreign import unsafe "glClearColor" glClearColor ::
-    GLclampf -> GLclampf -> GLclampf -> GLclampf -> IO ()
-
 --------------------------------------------------------------------------------
 
 -- | Controls the index /c/ used by 'clear' to clear the color index buffers.
@@ -346,8 +321,6 @@ clearIndex :: StateVar (Index1 GLfloat)
 clearIndex = makeStateVar (getFloat1 Index1 GetIndexClearValue)
                           (\(Index1 i) -> glClearIndex i)
 
-foreign import unsafe "glClearIndex" glClearIndex :: GLfloat -> IO ()
-
 --------------------------------------------------------------------------------
 
 -- | Controls the depth value used by 'clear' to clear the depth buffer. Values
@@ -355,9 +328,7 @@ foreign import unsafe "glClearIndex" glClearIndex :: GLfloat -> IO ()
 -- is 1.
 
 clearDepth :: StateVar GLclampd
-clearDepth = makeStateVar (getDouble1 id GetDepthClearValue) glClearDepth
-
-foreign import unsafe "glClearDepth" glClearDepth :: GLclampd -> IO ()
+clearDepth = makeStateVar (getClampd1 id GetDepthClearValue) glClearDepth
 
 --------------------------------------------------------------------------------
 
@@ -368,8 +339,6 @@ foreign import unsafe "glClearDepth" glClearDepth :: GLclampd -> IO ()
 clearStencil :: StateVar GLint
 clearStencil = makeStateVar (getInteger1 id GetStencilClearValue) glClearStencil
 
-foreign import unsafe "glClearStencil" glClearStencil :: GLint -> IO ()
-
 --------------------------------------------------------------------------------
 
 -- | Controls the red, green, blue, and alpha values used by 'clear' to clear
@@ -381,9 +350,6 @@ clearAccum =
    makeStateVar (getFloat4 Color4 GetAccumClearValue)
                 (\(Color4 r g b a) -> glClearAccum r g b a)
 
-foreign import unsafe "glClearAccum" glClearAccum ::
-   GLfloat -> GLfloat -> GLfloat -> GLfloat -> IO ()
-
 --------------------------------------------------------------------------------
 
 -- | An operation on the accumulation buffer.
@@ -425,11 +391,11 @@ data AccumOp =
 
 marshalAccumOp :: AccumOp -> GLenum
 marshalAccumOp x = case x of
-   Accum -> 0x100
-   Load -> 0x101
-   Return -> 0x102
-   Mult -> 0x103
-   Add -> 0x104
+   Accum -> gl_ACCUM
+   Load -> gl_LOAD
+   Return -> gl_RETURN
+   Mult -> gl_MULT
+   Add -> gl_ADD
 
 --------------------------------------------------------------------------------
 
@@ -461,5 +427,3 @@ marshalAccumOp x = case x of
 
 accum :: AccumOp -> GLfloat -> IO ()
 accum = glAccum . marshalAccumOp
-
-foreign import unsafe "glAccum" glAccum :: GLenum -> GLfloat -> IO ()
