@@ -1,16 +1,12 @@
-# 1 "VertexArrays.hs_pre"
-# 1 "<built-in>"
-# 1 "<command-line>"
-# 1 "VertexArrays.hs_pre"
 --------------------------------------------------------------------------------
 -- |
--- Module : Graphics.Rendering.OpenGL.GL.VertexArrays
--- Copyright : (c) Sven Panne 2002-2006
--- License : BSD-style (see the file libraries/OpenGL/LICENSE)
+-- Module      :  Graphics.Rendering.OpenGL.GL.VertexArrays
+-- Copyright   :  (c) Sven Panne 2002-2013
+-- License     :  BSD3
 --
--- Maintainer : sven.panne@aedion.de
--- Stability : stable
--- Portability : portable
+-- Maintainer  :  Sven Panne <svenpanne@gmail.com>
+-- Stability   :  stable
+-- Portability :  portable
 --
 -- This module corresponds to section 2.8 (Vertex Arrays) of the OpenGL 2.1
 -- specs.
@@ -22,6 +18,7 @@ module Graphics.Rendering.OpenGL.GL.VertexArrays (
    NumComponents, DataType(..), Stride, VertexArrayDescriptor(..),
 
    -- * Specifying Data for the Arrays
+   Capability(..),
    ClientArrayType(..), arrayPointer,
    InterleavedArrays(..), interleavedArrays,
 
@@ -32,50 +29,25 @@ module Graphics.Rendering.OpenGL.GL.VertexArrays (
    ArrayIndex, NumArrayIndices, NumIndexBlocks,
    arrayElement, drawArrays, multiDrawArrays, drawElements, multiDrawElements,
    drawRangeElements, maxElementsVertices, maxElementsIndices, lockArrays,
-   primitiveRestartIndex
+   primitiveRestartIndex, primitiveRestartIndexNV,
+
+   -- * Generic Vertex Attribute Arrays
+   vertexAttribPointer, vertexAttribArray,
 ) where
 
-import Foreign.Marshal.Alloc ( alloca )
-import Foreign.Ptr ( Ptr, nullPtr )
-import Foreign.Storable ( Storable(peek) )
-import Graphics.Rendering.OpenGL.GL.Capability (
-   EnableCap(CapVertexArray,CapNormalArray,CapColorArray,CapIndexArray,
-             CapTextureCoordArray,CapEdgeFlagArray,CapFogCoordArray,
-             CapSecondaryColorArray,CapMatrixIndexArray,CapPrimitiveRestart),
-   makeCapability )
-import Graphics.Rendering.OpenGL.GL.BasicTypes (
-   GLboolean, GLenum, GLint, GLuint, GLsizei, Capability(Enabled) )
-import Graphics.Rendering.OpenGL.GL.DataType (
-   DataType(..), marshalDataType, unmarshalDataType )
-import Graphics.Rendering.OpenGL.GL.Extensions (
-   FunPtr, unsafePerformIO, Invoker, getProcAddress )
-import Graphics.Rendering.OpenGL.GL.QueryUtils (
-   GetPName(GetVertexArraySize,GetVertexArrayType,GetVertexArrayStride,
-            GetNormalArrayType,GetNormalArrayStride,GetColorArraySize,
-            GetColorArrayType,GetColorArrayStride,GetSecondaryColorArraySize,
-            GetSecondaryColorArrayType,GetSecondaryColorArrayStride,
-            GetIndexArrayType,GetIndexArrayStride,
-            GetFogCoordArrayType,GetFogCoordArrayStride,
-            GetTextureCoordArraySize,GetTextureCoordArrayType,
-            GetTextureCoordArrayStride,GetEdgeFlagArrayStride,
-            GetMaxElementsVertices,GetMaxElementsIndices,
-            GetClientActiveTexture,GetArrayElementLockFirst,
-            GetArrayElementLockCount,GetPrimitiveRestartIndex),
-   getInteger1, getEnum1, getSizei1 )
-import Graphics.Rendering.OpenGL.GL.PrimitiveMode ( marshalPrimitiveMode )
-import Graphics.Rendering.OpenGL.GL.BeginEnd ( PrimitiveMode )
-import Graphics.Rendering.OpenGL.GL.StateVar (
-   HasGetter(get),
-   GettableStateVar, makeGettableStateVar, StateVar, makeStateVar )
-import Graphics.Rendering.OpenGL.GL.Texturing.TextureUnit (
-   TextureUnit, marshalTextureUnit, unmarshalTextureUnit )
-import Graphics.Rendering.OpenGL.GLU.ErrorsInternal (
-   recordInvalidEnum, recordInvalidValue )
-
---------------------------------------------------------------------------------
-
-# 1 "../../../../include/HsOpenGLExt.h" 1
-# 74 "VertexArrays.hs_pre" 2
+import Foreign.Marshal.Alloc
+import Foreign.Ptr
+import Foreign.Storable
+import Graphics.Rendering.OpenGL.GL.Capability
+import Graphics.Rendering.OpenGL.GL.DataType
+import Graphics.Rendering.OpenGL.GL.GLboolean
+import Graphics.Rendering.OpenGL.GL.PrimitiveMode
+import Graphics.Rendering.OpenGL.GL.QueryUtils
+import Graphics.Rendering.OpenGL.GL.StateVar
+import Graphics.Rendering.OpenGL.GL.Texturing.TextureUnit
+import Graphics.Rendering.OpenGL.GL.VertexSpec
+import Graphics.Rendering.OpenGL.GLU.ErrorsInternal
+import Graphics.Rendering.OpenGL.Raw
 
 --------------------------------------------------------------------------------
 
@@ -85,15 +57,8 @@ type Stride = GLsizei
 
 data VertexArrayDescriptor a =
    VertexArrayDescriptor !NumComponents !DataType !Stride !(Ptr a)
-
-
-
-
-
-
    deriving ( Eq, Ord, Show )
 
-
 noVertexArrayDescriptor :: VertexArrayDescriptor a
 noVertexArrayDescriptor = VertexArrayDescriptor 0 Byte 0 nullPtr
 
@@ -113,15 +78,15 @@ data ClientArrayType =
 
 marshalClientArrayType :: ClientArrayType -> GLenum
 marshalClientArrayType x = case x of
-   VertexArray -> 0x8074
-   NormalArray -> 0x8075
-   ColorArray -> 0x8076
-   IndexArray -> 0x8077
-   TextureCoordArray -> 0x8078
-   EdgeFlagArray -> 0x8079
-   FogCoordArray -> 0x8457
-   SecondaryColorArray -> 0x845e
-   MatrixIndexArray -> 0x8844
+   VertexArray -> gl_VERTEX_ARRAY
+   NormalArray -> gl_NORMAL_ARRAY
+   ColorArray -> gl_COLOR_ARRAY
+   IndexArray -> gl_INDEX_ARRAY
+   TextureCoordArray -> gl_TEXTURE_COORD_ARRAY
+   EdgeFlagArray -> gl_EDGE_FLAG_ARRAY
+   FogCoordArray -> gl_FOG_COORD_ARRAY
+   SecondaryColorArray -> gl_SECONDARY_COLOR_ARRAY
+   MatrixIndexArray -> gl_MATRIX_INDEX_ARRAY
 
 -- Hmmm...
 clientArrayTypeToEnableCap :: ClientArrayType -> EnableCap
@@ -173,9 +138,6 @@ setVertexPointer :: VertexArrayDescriptor a -> IO ()
 setVertexPointer (VertexArrayDescriptor n d s p) =
    glVertexPointer n (marshalDataType d) s p
 
-foreign import unsafe "glVertexPointer" glVertexPointer ::
-   GLint -> GLenum -> GLsizei -> Ptr a -> IO ()
-
 --------------------------------------------------------------------------------
 
 normalPointer :: StateVar (VertexArrayDescriptor a)
@@ -192,9 +154,6 @@ setNormalPointer :: VertexArrayDescriptor a -> IO ()
 setNormalPointer (VertexArrayDescriptor n d s p) =
    check (n == 3) $ glNormalPointer (marshalDataType d) s p
 
-foreign import unsafe "glNormalPointer" glNormalPointer ::
-   GLenum -> GLsizei -> Ptr a -> IO ()
-
 --------------------------------------------------------------------------------
 
 colorPointer :: StateVar (VertexArrayDescriptor a)
@@ -212,9 +171,6 @@ setColorPointer :: VertexArrayDescriptor a -> IO ()
 setColorPointer (VertexArrayDescriptor n d s p) =
    check (n == 3 || n == 4) $ glColorPointer n (marshalDataType d) s p
 
-foreign import unsafe "glColorPointer" glColorPointer ::
-   GLint -> GLenum -> GLsizei -> Ptr a -> IO ()
-
 --------------------------------------------------------------------------------
 
 indexPointer :: StateVar (VertexArrayDescriptor a)
@@ -231,9 +187,6 @@ setIndexPointer :: VertexArrayDescriptor a -> IO ()
 setIndexPointer (VertexArrayDescriptor n d s p) =
    check (n == 1) $ glIndexPointer (marshalDataType d) s p
 
-foreign import unsafe "glIndexPointer" glIndexPointer ::
-   GLenum -> GLsizei -> Ptr a -> IO ()
-
 --------------------------------------------------------------------------------
 
 texCoordPointer :: StateVar (VertexArrayDescriptor a)
@@ -251,9 +204,6 @@ setTexCoordPointer :: VertexArrayDescriptor a -> IO ()
 setTexCoordPointer (VertexArrayDescriptor n d s p) =
    glTexCoordPointer n (marshalDataType d) s p
 
-foreign import unsafe "glTexCoordPointer" glTexCoordPointer ::
-   GLint -> GLenum -> GLsizei -> Ptr a -> IO ()
-
 --------------------------------------------------------------------------------
 
 edgeFlagPointer :: StateVar (VertexArrayDescriptor a)
@@ -269,9 +219,6 @@ setEdgeFlagPointer :: VertexArrayDescriptor a -> IO ()
 setEdgeFlagPointer (VertexArrayDescriptor n d s p) =
    check (n == 1 && d == UnsignedByte) $ glEdgeFlagPointer s p
 
-foreign import unsafe "glEdgeFlagPointer" glEdgeFlagPointer ::
-   GLsizei -> Ptr a -> IO ()
-
 --------------------------------------------------------------------------------
 
 fogCoordPointer :: StateVar (VertexArrayDescriptor a)
@@ -286,9 +233,7 @@ getFogCoordPointer = do
 
 setFogCoordPointer :: VertexArrayDescriptor a -> IO ()
 setFogCoordPointer (VertexArrayDescriptor n d s p) =
-   check (n == 1) $ glFogCoordPointerEXT (marshalDataType d) s p
-
-foreign import unsafe "dynamic" dyn_glFogCoordPointerEXT :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLenum -> GLsizei -> Ptr a -> IO ()) ; glFogCoordPointerEXT :: (GLenum -> GLsizei -> Ptr a -> IO ()) ; glFogCoordPointerEXT = dyn_glFogCoordPointerEXT ptr_glFogCoordPointerEXT ; ptr_glFogCoordPointerEXT :: FunPtr a ; ptr_glFogCoordPointerEXT = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_EXT_fog_coord or OpenGL 1.4") ("_entry")) ;
+   check (n == 1) $ glFogCoordPointer (marshalDataType d) s p
 
 --------------------------------------------------------------------------------
 
@@ -306,16 +251,7 @@ getSecondaryColorPointer = do
 
 setSecondaryColorPointer :: (VertexArrayDescriptor a) -> IO ()
 setSecondaryColorPointer (VertexArrayDescriptor n d s p) =
-   glSecondaryColorPointerEXT n (marshalDataType d) s p
-
-foreign import unsafe "dynamic" dyn_glSecondaryColorPointerEXT :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLint -> GLenum -> GLsizei -> Ptr a -> IO ()) ; glSecondaryColorPointerEXT :: (GLint -> GLenum -> GLsizei -> Ptr a -> IO ()) ; glSecondaryColorPointerEXT = dyn_glSecondaryColorPointerEXT ptr_glSecondaryColorPointerEXT ; ptr_glSecondaryColorPointerEXT :: FunPtr a ; ptr_glSecondaryColorPointerEXT = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_EXT_secondary_color or OpenGL 1.4") ("_entry")) ;
-
---------------------------------------------------------------------------------
-
-foreign import unsafe "dynamic" dyn_glVertexAttribPointer :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLuint -> GLint -> GLenum -> GLboolean -> GLsizei -> Ptr a -> IO ()) ; glVertexAttribPointer :: (GLuint -> GLint -> GLenum -> GLboolean -> GLsizei -> Ptr a -> IO ()) ; glVertexAttribPointer = dyn_glVertexAttribPointer ptr_glVertexAttribPointer ; ptr_glVertexAttribPointer :: FunPtr a ; ptr_glVertexAttribPointer = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("OpenGL 2.0") ("_entry")) ;
-foreign import unsafe "dynamic" dyn_glDisableVertexAttribArray :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLuint -> IO ()) ; glDisableVertexAttribArray :: (GLuint -> IO ()) ; glDisableVertexAttribArray = dyn_glDisableVertexAttribArray ptr_glDisableVertexAttribArray ; ptr_glDisableVertexAttribArray :: FunPtr a ; ptr_glDisableVertexAttribArray = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("OpenGL 2.0") ("_entry")) ;
-foreign import unsafe "dynamic" dyn_glEnableVertexAttribArray :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLuint -> IO ()) ; glEnableVertexAttribArray :: (GLuint -> IO ()) ; glEnableVertexAttribArray = dyn_glEnableVertexAttribArray ptr_glEnableVertexAttribArray ; ptr_glEnableVertexAttribArray :: FunPtr a ; ptr_glEnableVertexAttribArray = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("OpenGL 2.0") ("_entry")) ;
-foreign import unsafe "dynamic" dyn_glGetVertexAttribPointerv :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLuint -> GLenum -> Ptr (Ptr a) -> IO ()) ; glGetVertexAttribPointerv :: (GLuint -> GLenum -> Ptr (Ptr a) -> IO ()) ; glGetVertexAttribPointerv = dyn_glGetVertexAttribPointerv ptr_glGetVertexAttribPointerv ; ptr_glGetVertexAttribPointerv :: FunPtr a ; ptr_glGetVertexAttribPointerv = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("OpenGL 2.0") ("_entry")) ;
+   glSecondaryColorPointer n (marshalDataType d) s p
 
 --------------------------------------------------------------------------------
 
@@ -338,29 +274,26 @@ data InterleavedArrays =
 
 marshalInterleavedArrays :: InterleavedArrays -> GLenum
 marshalInterleavedArrays x = case x of
-   V2f -> 0x2a20
-   V3f -> 0x2a21
-   C4ubV2f -> 0x2a22
-   C4ubV3f -> 0x2a23
-   C3fV3f -> 0x2a24
-   N3fV3f -> 0x2a25
-   C4fN3fV3f -> 0x2a26
-   T2fV3f -> 0x2a27
-   T4fV4f -> 0x2a28
-   T2fC4ubV3f -> 0x2a29
-   T2fC3fV3f -> 0x2a2a
-   T2fN3fV3f -> 0x2a2b
-   T2fC4fN3fV3f -> 0x2a2c
-   T4fC4fN3fV4f -> 0x2a2d
+   V2f -> gl_V2F
+   V3f -> gl_V3F
+   C4ubV2f -> gl_C4UB_V2F
+   C4ubV3f -> gl_C4UB_V3F
+   C3fV3f -> gl_C3F_V3F
+   N3fV3f -> gl_N3F_V3F
+   C4fN3fV3f -> gl_C4F_N3F_V3F
+   T2fV3f -> gl_T2F_V3F
+   T4fV4f -> gl_T4F_V4F
+   T2fC4ubV3f -> gl_T2F_C4UB_V3F
+   T2fC3fV3f -> gl_T2F_C3F_V3F
+   T2fN3fV3f -> gl_T2F_N3F_V3F
+   T2fC4fN3fV3f -> gl_T2F_C4F_N3F_V3F
+   T4fC4fN3fV4f -> gl_T4F_C4F_N3F_V4F
 
 --------------------------------------------------------------------------------
 
 interleavedArrays :: InterleavedArrays -> Stride -> Ptr a -> IO ()
 interleavedArrays = glInterleavedArrays . marshalInterleavedArrays
 
-foreign import unsafe "glInterleavedArrays" glInterleavedArrays ::
-   GLenum -> GLsizei -> Ptr a -> IO ()
-
 --------------------------------------------------------------------------------
 
 clientState :: ClientArrayType -> StateVar Capability
@@ -368,27 +301,19 @@ clientState arrayType =
    makeStateVar (getClientState arrayType) (setClientState arrayType)
 
 getClientState :: ClientArrayType -> IO Capability
-getClientState = get . makeCapability . clientArrayTypeToEnableCap
+getClientState arrayType = get . makeCapability . clientArrayTypeToEnableCap $ arrayType
 
 setClientState :: ClientArrayType -> Capability -> IO ()
 setClientState arrayType val =
    (if val == Enabled then glEnableClientState else glDisableClientState)
       (marshalClientArrayType arrayType)
 
-foreign import unsafe "glEnableClientState" glEnableClientState ::
-   GLenum -> IO ()
-
-foreign import unsafe "glDisableClientState" glDisableClientState ::
-   GLenum -> IO ()
-
 --------------------------------------------------------------------------------
 
 clientActiveTexture :: StateVar TextureUnit
 clientActiveTexture =
    makeStateVar (getEnum1 unmarshalTextureUnit GetClientActiveTexture)
-                (glClientActiveTextureARB . marshalTextureUnit)
-
-foreign import unsafe "dynamic" dyn_glClientActiveTextureARB :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLenum -> IO ()) ; glClientActiveTextureARB :: (GLenum -> IO ()) ; glClientActiveTextureARB = dyn_glClientActiveTextureARB ptr_glClientActiveTextureARB ; ptr_glClientActiveTextureARB :: FunPtr a ; ptr_glClientActiveTextureARB = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_ARB_multitexture or OpenGL 1.3") ("_entry")) ;
+                (glClientActiveTexture . marshalTextureUnit)
 
 --------------------------------------------------------------------------------
 
@@ -403,43 +328,29 @@ type NumIndexBlocks = GLsizei
 arrayElement :: ArrayIndex -> IO ()
 arrayElement = glArrayElement
 
-foreign import unsafe "glArrayElement" glArrayElement :: GLint -> IO ()
-
 drawArrays :: PrimitiveMode -> ArrayIndex -> NumArrayIndices -> IO ()
 drawArrays = glDrawArrays . marshalPrimitiveMode
 
-foreign import unsafe "glDrawArrays" glDrawArrays ::
-   GLenum -> GLint -> GLsizei -> IO ()
-
 multiDrawArrays ::
       PrimitiveMode -> Ptr ArrayIndex -> Ptr NumArrayIndices -> NumIndexBlocks
    -> IO ()
-multiDrawArrays = glMultiDrawArraysEXT . marshalPrimitiveMode
-
-foreign import unsafe "dynamic" dyn_glMultiDrawArraysEXT :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLenum -> Ptr ArrayIndex -> Ptr GLsizei -> GLsizei -> IO ()) ; glMultiDrawArraysEXT :: (GLenum -> Ptr ArrayIndex -> Ptr GLsizei -> GLsizei -> IO ()) ; glMultiDrawArraysEXT = dyn_glMultiDrawArraysEXT ptr_glMultiDrawArraysEXT ; ptr_glMultiDrawArraysEXT :: FunPtr a ; ptr_glMultiDrawArraysEXT = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_EXT_multi_draw_arrays or OpenGL 1.4") ("_entry")) ;
+multiDrawArrays = glMultiDrawArrays . marshalPrimitiveMode
 
 drawElements :: PrimitiveMode -> NumArrayIndices -> DataType -> Ptr a -> IO ()
 drawElements m c = glDrawElements (marshalPrimitiveMode m) c . marshalDataType
 
-foreign import unsafe "glDrawElements" glDrawElements ::
-   GLenum -> GLsizei -> GLenum -> Ptr a -> IO ()
-
 multiDrawElements ::
       PrimitiveMode -> Ptr NumArrayIndices -> DataType -> Ptr (Ptr a)
    -> NumIndexBlocks -> IO ()
 multiDrawElements m c =
-   glMultiDrawElementsEXT (marshalPrimitiveMode m) c . marshalDataType
-
-foreign import unsafe "dynamic" dyn_glMultiDrawElementsEXT :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLenum -> Ptr GLsizei -> GLenum -> Ptr (Ptr a) -> GLsizei -> IO ()) ; glMultiDrawElementsEXT :: (GLenum -> Ptr GLsizei -> GLenum -> Ptr (Ptr a) -> GLsizei -> IO ()) ; glMultiDrawElementsEXT = dyn_glMultiDrawElementsEXT ptr_glMultiDrawElementsEXT ; ptr_glMultiDrawElementsEXT :: FunPtr a ; ptr_glMultiDrawElementsEXT = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_EXT_multi_draw_arrays or OpenGL 1.4") ("_entry")) ;
+   glMultiDrawElements (marshalPrimitiveMode m) c . marshalDataType
 
 drawRangeElements ::
       PrimitiveMode -> (ArrayIndex, ArrayIndex) -> NumArrayIndices -> DataType
    -> Ptr a -> IO ()
 drawRangeElements m (s, e) c =
-   glDrawRangeElementsEXT (marshalPrimitiveMode m) (fromIntegral s)
-                          (fromIntegral e) c . marshalDataType
-
-foreign import unsafe "dynamic" dyn_glDrawRangeElementsEXT :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLenum -> GLuint -> GLuint -> GLsizei -> GLenum -> Ptr a -> IO ()) ; glDrawRangeElementsEXT :: (GLenum -> GLuint -> GLuint -> GLsizei -> GLenum -> Ptr a -> IO ()) ; glDrawRangeElementsEXT = dyn_glDrawRangeElementsEXT ptr_glDrawRangeElementsEXT ; ptr_glDrawRangeElementsEXT :: FunPtr a ; ptr_glDrawRangeElementsEXT = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_EXT_draw_range_elements or OpenGL 1.2") ("_entry")) ;
+   glDrawRangeElements (marshalPrimitiveMode m) (fromIntegral s)
+                       (fromIntegral e) c . marshalDataType
 
 maxElementsVertices :: GettableStateVar NumArrayIndices
 maxElementsVertices = makeGettableStateVar (getSizei1 id GetMaxElementsVertices)
@@ -461,34 +372,37 @@ getLockArrays = do
       else return Nothing
 
 setLockArrays :: Maybe (ArrayIndex, NumArrayIndices) -> IO ()
-setLockArrays = maybe glUnlockArraysEXT (uncurry glLockArraysEXT)
+setLockArrays = maybe glUnlockArrays (uncurry glLockArrays)
 
-foreign import unsafe "dynamic" dyn_glLockArraysEXT :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLint -> GLsizei -> IO ()) ; glLockArraysEXT :: (GLint -> GLsizei -> IO ()) ; glLockArraysEXT = dyn_glLockArraysEXT ptr_glLockArraysEXT ; ptr_glLockArraysEXT :: FunPtr a ; ptr_glLockArraysEXT = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_EXT_compiled_vertex_array") ("_entry")) ;
-foreign import unsafe "dynamic" dyn_glUnlockArraysEXT :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (IO ()) ; glUnlockArraysEXT :: (IO ()) ; glUnlockArraysEXT = dyn_glUnlockArraysEXT ptr_glUnlockArraysEXT ; ptr_glUnlockArraysEXT :: FunPtr a ; ptr_glUnlockArraysEXT = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_EXT_compiled_vertex_array") ("_entry")) ;
+--------------------------------------------------------------------------------
+
+primitiveRestartIndex :: StateVar (Maybe ArrayIndex)
+primitiveRestartIndex =
+   makeStateVarMaybe
+      (return CapPrimitiveRestart)
+      (getInteger1 id GetPrimitiveRestartIndex)
+      (glPrimitiveRestartIndex . fromIntegral)
 
 --------------------------------------------------------------------------------
 
 -- We almost could use makeStateVarMaybe below, but, alas, this is client state.
 
-primitiveRestartIndex :: StateVar (Maybe ArrayIndex)
-primitiveRestartIndex =
-   makeStateVar getPrimitiverestartIndex setPrimitiverestartIndex
+primitiveRestartIndexNV :: StateVar (Maybe ArrayIndex)
+primitiveRestartIndexNV =
+   makeStateVar getPrimitiveRestartIndexNV setPrimitiveRestartIndexNV
 
-getPrimitiverestartIndex :: IO (Maybe ArrayIndex)
-getPrimitiverestartIndex = do
-   state <- get (makeCapability CapPrimitiveRestart)
-   if state == Enabled
-      then fmap Just $ getInteger1 fromIntegral GetPrimitiveRestartIndex
+getPrimitiveRestartIndexNV :: IO (Maybe ArrayIndex)
+getPrimitiveRestartIndexNV = do
+   on <- getBoolean1 unmarshalGLboolean GetPrimitiveRestartNV
+   if on
+      then fmap Just $ getInteger1 fromIntegral GetPrimitiveRestartIndexNV
       else return Nothing
 
-setPrimitiverestartIndex :: Maybe ArrayIndex -> IO ()
-setPrimitiverestartIndex maybeIdx = case maybeIdx of
-   Nothing -> glDisableClientState primitiveRestartNV
-   Just idx -> do glEnableClientState primitiveRestartNV
+setPrimitiveRestartIndexNV :: Maybe ArrayIndex -> IO ()
+setPrimitiveRestartIndexNV maybeIdx = case maybeIdx of
+   Nothing  -> glDisableClientState gl_PRIMITIVE_RESTART_NV
+   Just idx -> do glEnableClientState gl_PRIMITIVE_RESTART_NV
                   glPrimitiveRestartIndexNV (fromIntegral idx)
-   where primitiveRestartNV = 0x8558 -- ToDo: HACK!
-
-foreign import unsafe "dynamic" dyn_glPrimitiveRestartIndexNV :: Graphics.Rendering.OpenGL.GL.Extensions.Invoker (GLuint -> IO ()) ; glPrimitiveRestartIndexNV :: (GLuint -> IO ()) ; glPrimitiveRestartIndexNV = dyn_glPrimitiveRestartIndexNV ptr_glPrimitiveRestartIndexNV ; ptr_glPrimitiveRestartIndexNV :: FunPtr a ; ptr_glPrimitiveRestartIndexNV = unsafePerformIO (Graphics.Rendering.OpenGL.GL.Extensions.getProcAddress ("GL_NV_primitive_restart") ("_entry")) ;
 
 --------------------------------------------------------------------------------
 
@@ -508,18 +422,18 @@ data GetPointervPName =
 
 marshalGetPointervPName :: GetPointervPName -> GLenum
 marshalGetPointervPName x = case x of
-   VertexArrayPointer -> 0x808e
-   NormalArrayPointer -> 0x808f
-   ColorArrayPointer -> 0x8090
-   IndexArrayPointer -> 0x8091
-   TextureCoordArrayPointer -> 0x8092
-   EdgeFlagArrayPointer -> 0x8093
-   FogCoordArrayPointer -> 0x8456
-   SecondaryColorArrayPointer -> 0x845d
-   FeedbackBufferPointer -> 0xdf0
-   SelectionBufferPointer -> 0xdf3
-   WeightArrayPointer -> 0x86ac
-   MatrixIndexArrayPointer -> 0x8849
+   VertexArrayPointer -> gl_VERTEX_ARRAY_POINTER
+   NormalArrayPointer -> gl_NORMAL_ARRAY_POINTER
+   ColorArrayPointer -> gl_COLOR_ARRAY_POINTER
+   IndexArrayPointer -> gl_INDEX_ARRAY_POINTER
+   TextureCoordArrayPointer -> gl_TEXTURE_COORD_ARRAY_POINTER
+   EdgeFlagArrayPointer -> gl_EDGE_FLAG_ARRAY_POINTER
+   FogCoordArrayPointer -> gl_FOG_COORD_ARRAY_POINTER
+   SecondaryColorArrayPointer -> gl_SECONDARY_COLOR_ARRAY_POINTER
+   FeedbackBufferPointer -> gl_FEEDBACK_BUFFER_POINTER
+   SelectionBufferPointer -> gl_SELECTION_BUFFER_POINTER
+   WeightArrayPointer -> gl_WEIGHT_ARRAY_POINTER
+   MatrixIndexArrayPointer -> gl_MATRIX_INDEX_ARRAY_POINTER
 
 --------------------------------------------------------------------------------
 
@@ -528,5 +442,42 @@ getPointer n = alloca $ \buf -> do
    glGetPointerv (marshalGetPointervPName n) buf
    peek buf
 
-foreign import unsafe "glGetPointerv" glGetPointerv ::
-   GLenum -> Ptr (Ptr a) -> IO ()
+--------------------------------------------------------------------------------
+
+vertexAttribPointer :: AttribLocation -> StateVar (IntegerHandling, VertexArrayDescriptor a)
+vertexAttribPointer location =
+   makeStateVar (getVertexAttribPointer_ location) (setVertexAttribPointer location)
+
+getVertexAttribPointer_ :: AttribLocation -> IO (IntegerHandling, VertexArrayDescriptor a)
+getVertexAttribPointer_ location = do
+   i <- getVertexAttribBoolean1 unmarshalGLboolean location GetVertexAttribArrayInteger
+   h <- if i
+           then return KeepIntegral
+           else do f <- getVertexAttribBoolean1 unmarshalGLboolean location GetVertexAttribArrayNormalized
+                   return $ if f then ToNormalizedFloat else ToFloat
+   n <- getVertexAttribInteger1 id location GetVertexAttribArraySize
+   d <- getVertexAttribEnum1 unmarshalDataType location GetVertexAttribArrayType
+   s <- getVertexAttribInteger1 fromIntegral location GetVertexAttribArrayStride
+   p <- getVertexAttribPointer location VertexAttribArrayPointer
+   return (h, VertexArrayDescriptor n d s p)
+
+setVertexAttribPointer :: AttribLocation -> (IntegerHandling, VertexArrayDescriptor a) -> IO ()
+setVertexAttribPointer (AttribLocation location) (h, VertexArrayDescriptor n d s p) = case h of
+   ToFloat -> glVertexAttribPointer location n md (marshalGLboolean False) s p
+   ToNormalizedFloat -> glVertexAttribPointer location n md (marshalGLboolean True) s p
+   KeepIntegral -> glVertexAttribIPointer location n md s p
+  where md = marshalDataType d
+
+--------------------------------------------------------------------------------
+
+vertexAttribArray :: AttribLocation -> StateVar Capability
+vertexAttribArray location =
+   makeStateVar (getVertexAttribArray location) (flip setVertexAttribArray location)
+
+getVertexAttribArray :: AttribLocation -> IO Capability
+getVertexAttribArray location =
+   getVertexAttribBoolean1 unmarshalCapability location GetVertexAttribArrayEnabled
+
+setVertexAttribArray :: Capability -> AttribLocation -> IO ()
+setVertexAttribArray Disabled (AttribLocation location) = glDisableVertexAttribArray location
+setVertexAttribArray Enabled (AttribLocation location) = glEnableVertexAttribArray location
