@@ -1,16 +1,12 @@
-# 1 "Errors.hs_pre"
-# 1 "<built-in>"
-# 1 "<command-line>"
-# 1 "Errors.hs_pre"
 --------------------------------------------------------------------------------
 -- |
--- Module : Graphics.Rendering.OpenGL.GLU.Errors
--- Copyright : (c) Sven Panne 2002-2006
--- License : BSD-style (see the file libraries/OpenGL/LICENSE)
---
--- Maintainer : sven.panne@aedion.de
--- Stability : stable
--- Portability : portable
+-- Module      :  Graphics.Rendering.OpenGL.GLU.Errors
+-- Copyright   :  (c) Sven Panne 2002-2013
+-- License     :  BSD3
+-- 
+-- Maintainer  :  Sven Panne <svenpanne@gmail.com>
+-- Stability   :  stable
+-- Portability :  portable
 --
 -- This module corresponds to section 2.5 (GL Errors) of the OpenGL 2.1 specs
 -- and chapter 8 (Errors) of the GLU specs, offering a generalized view of
@@ -22,10 +18,8 @@ module Graphics.Rendering.OpenGL.GLU.Errors (
    Error(..), ErrorCategory(..), errors
 ) where
 
-import Graphics.Rendering.OpenGL.GL.StateVar (
-   GettableStateVar, makeGettableStateVar )
-import Graphics.Rendering.OpenGL.GLU.ErrorsInternal (
-   Error(..), ErrorCategory(..), getErrors )
+import Graphics.Rendering.OpenGL.GL.StateVar
+import Graphics.Rendering.OpenGL.GLU.ErrorsInternal
 
 --------------------------------------------------------------------------------
 
